import asyncio
from typing import Any, Dict


def __init__(hub):
    # A mapping of dyne_names/subs to a reference to a function on the hub that should run that function
    # If no reference is listed then a getattr(hub, ref) will be done and the reference will be called directly
    hub.gate.srv.RUNNER = {}


async def runner(hub, ref: str, params: Dict[str, Any]) -> Any:
    """
    Call a function based on runners defined before the program started
    """
    prefix = ref.split(".")[0]
    if prefix in hub.gate.srv.RUNNER:
        run = hub.gate.srv.RUNNER[prefix]
        if not hasattr(run, "__call__"):
            # If the prefix was a string then get it's reference on the hub
            run = getattr(hub, run)
        ret = run(ref, **params)
    else:
        ret = getattr(hub, ref)(**params)

    if asyncio.iscoroutine(ret):
        ret = await ret

    return ret



from typing import Dict


def match(hub, ref: str) -> bool:
    """
    Match the listed refs with the desired ref to see if the correct ref is
    being exposed.
    """
    return hub.gate.matcher[hub.OPT.gate.matcher].match(ref)



CLI_CONFIG = {
    "host": {},
    "port": {},
    "server": {},
    "prefix": {},
    "matcher": {},
    "refs": {},
}
CONFIG = {
    "host": {
        "default": "0.0.0.0",
        "type": str,
        "help": "The host interface to bind to.",
    },
    "port": {"default": 8080, "type": int, "help": "The port to bind to"},
    "server": {"default": "starlette", "help": "The server plugin interface to use"},
    "prefix": {
        "default": None,
        "type": str,
        "help": "The prefix for hub references that can be exposed",
    },
    "matcher": {
        "default": "glob",
        "type": str,
        "help": "The matcher plugin that will be used on the refs",
    },
    "refs": {
        "default": ["gate.init.test"],
        "help": "The hub references to expose",
        "nargs": "*",
    },
}
SUBCOMMANDS = {}
DYNE = {
    "gate": ["gate"],
}



def __init__(hub):
    hub.pop.sub.load_subdirs(hub.gate, recurse=True)


def cli(hub):
    hub.pop.config.load(["gate"], cli="gate")
    hub.pop.loop.create()
    hub.pop.Loop.run_until_complete(
        hub.gate.init.start(gate_server=hub.OPT.gate.server)
    )


async def test(hub, *args, **kwargs):
    """
    The test function for gate
    """
    return {"args": args, "kwargs": kwargs}


async def start(hub, gate_server: str):
    return await hub.gate.srv[gate_server].start()


async def stop(hub, gate_server: str):
    hub.log.debug("Shutting down gate server")
    return await hub.gate.srv[gate_server].stop()


async def join(hub, gate_server):
    return await hub.gate.srv[gate_server].join()



# Import python libs
import asyncio

try:
    # Import third party libs
    from starlette.applications import Starlette
    from starlette.responses import JSONResponse
    from starlette.routing import Route
    from starlette.requests import Request
    from uvicorn import Server
    from uvicorn.config import Config

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


async def start(hub):
    async def _query(request):
        return await hub.gate.srv.starlette.query(request)

    routes = [
        Route("/", _query, methods=["GET", "POST", "PUT"]),
    ]
    app = Starlette(debug=True, routes=routes)
    config = Config(app, host=hub.OPT.gate.host, port=hub.OPT.gate.port, loop="none")
    hub.gate.srv.starlette.SERVER = Server(config)

    return await hub.gate.srv.starlette.SERVER.serve()


async def join(hub):
    """
    Block until the server has started
    """
    while not hasattr(hub.gate.srv.starlette, "SERVER"):
        await asyncio.sleep(0, loop=hub.pop.Loop)

    while not getattr(hub.gate.srv.starlette.SERVER, "started", False):
        await asyncio.sleep(0, loop=hub.pop.Loop)

    # Wait just one more cycle for everything to start
    await asyncio.sleep(0, loop=hub.pop.Loop)


async def stop(hub):
    hub.gate.srv.starlette.SERVER.should_exit = True
    hub.gate.srv.starlette.SERVER.force_exit = True
    await hub.gate.srv.starlette.SERVER.shutdown()


async def query(hub, request: Request) -> JSONResponse:
    if request.headers["Content-type"] == "application/json":
        q_params = await request.json()
    else:
        q_params = dict(request.query_params)

    if "ref" not in q_params:
        return JSONResponse(
            {"error": "Required ref not found in params"}, status_code=417
        )
    q_ref = q_params.pop("ref")
    if hub.OPT.gate.prefix:
        q_ref = f"{hub.OPT.gate.prefix}.{q_ref}"

    if not hub.gate.matcher.init.match(q_ref):
        return JSONResponse(
            {"error": f"The provided ref {q_ref} is not available for execution"},
            status_code=405,
        )

    if not hasattr(hub, q_ref):
        return JSONResponse(
            {"error": f"The provided ref {q_ref} does not exist"}, status_code=404
        )

    try:
        ret = await hub.gate.srv.init.runner(q_ref, q_params)
        return JSONResponse(ret, status_code=200)
    except Exception as e:
        return JSONResponse(
            {"error": f"The provided ref {q_ref} returned an exception: {str(e)}"},
            status_code=500,
        )



# Import python libs
import fnmatch


def match(hub, ref: str) -> bool:
    """
    Check the desired ref to see if it matches a globular ref in the refs list
    """
    for pattern in hub.OPT.gate.refs:
        if hub.OPT.gate.prefix:
            pattern = f"{hub.OPT.gate.prefix}.{pattern}"
        if fnmatch.fnmatch(ref, pattern):
            return True
    else:
        return False
